/*
    This sever, along with its client, uses "message passing"
    (over a socket) to implement a simple synchronization pattern
    between two processes. The synchronization pattern is to keep
    the two processes "alternating" with each other. Each process
    does a sequence of "computations," where each computation can
    take a different amount of time, and the processes are
    synchronized so that each process begins a computation exactly
    when the other process ends a computation.

    This server is based on the server-template.c program. The only
    part that has been changed is the section where the application
    level protocol gets implemented.
*/
#include <stdio.h>      // for fprintf() and fdopen()
#include <stdlib.h>     // for atoi() and exit()
#include <string.h>     // for memset(), bzero(), bcopy(), and strerror()
#include <errno.h>      // for errno

#include <unistd.h>     // for close()
#include <sys/socket.h> // for socket(), connect(), AF_INET, SOCK_STREAM
#include <arpa/inet.h>  // for sockaddr_in, htons(), htonl()
#include <netdb.h>      // for gethostbyaddr(), hostent, hstrerror(), and h_errno

// prototypes for two error handling functions (defined below)
void unix_error(char *msg);
void  dns_error(char *msg);

#define LISTENQUEUE  8  // second argument to listen()

int main(int argc, char **argv)
{
   int socket_fd;                   // listening socket file descriptor
   int connected_fd;                // connected socket file descriptor
   int optionvalue = 1;             // used in setsockopt()
   struct sockaddr_in localAddr;    // server's address structure
   struct sockaddr_in clientAddr;   // clients's address structure
   struct hostent *hp;              // used to get client's IP address
   int clientLength;                // needed for accept()
   char *h_addressp;
   FILE *connected_fpout;           // connected socket file pointer (i.e., stream)
   FILE *connected_fpin;            // connected socket file pointer

   int workTime;
   int i;

   if (argc != 2)
   {
      fprintf(stderr, "usage: %s <port>\n", argv[0]);
      exit(0);
   }

   // Create a reliable, stream socket using TCP.
   if ((socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
      unix_error("socket() error");


   // Eliminate "Address already in use" error from bind().
   if ( setsockopt(socket_fd, SOL_SOCKET, SO_REUSEADDR,
                   (const void*)&optionvalue, sizeof(int)) < 0 )
      return -1;


   // Construct a local address structure.
   memset(&localAddr, 0, sizeof(localAddr));      // zero out the address structure
   localAddr.sin_family = AF_INET;                // Internet address family
   localAddr.sin_addr.s_addr = htonl(INADDR_ANY); // any incoming interface
   localAddr.sin_port = htons( atoi(argv[1]) );   // fill in the port number


   // Bind to the local address
   if (bind(socket_fd, (struct sockaddr*)&localAddr, sizeof(localAddr)) < 0)
      unix_error("bind() error");


   // Make it a listening socket ready to accept connection requests
   if (listen(socket_fd, LISTENQUEUE) < 0)
      unix_error("listen() error");

   while(1)
   {
      clientLength = sizeof(clientAddr);  // needed by accept()

      // Wait for a connection from a client.
      connected_fd = accept(socket_fd, (struct sockaddr*)&clientAddr, &clientLength);
      if (connected_fd < 0)
         unix_error("accept() error");


      // Determine the domain name and IP address of the client.
      hp = gethostbyaddr( (const char*)&clientAddr.sin_addr.s_addr,
                           sizeof(clientAddr.sin_addr.s_addr), AF_INET );
      if ( hp == NULL )
         dns_error("gethostbyaddr() error");
      h_addressp = inet_ntoa(clientAddr.sin_addr);
      fprintf(stderr, "server connected to %s (%s)\n", hp->h_name, h_addressp);


      // Convert the connected socket file descriptor to file pointers
      // (i.e., streams) and use C Standard Library I/O functions.
      if ( (connected_fpout = fdopen(connected_fd, "w")) == NULL )
         unix_error("fdopen() error" );
      if ( (connected_fpin = fdopen(connected_fd, "r")) == NULL )
         unix_error("fdopen() error" );


      while(1) // loop until client gives up
      {
         fscanf(connected_fpin, "%d", &workTime); // wait for the client
         if (workTime == -1) break;               // finished with this client
         workTime = 1 + rand()%9;                 // 1 to 9 seconds of work time
         fprintf(stderr, "\nServer: Starting a %d second calculation", workTime);
         for (i = 0; i < workTime; i++)
         {
            fprintf(stderr, ".");
            sleep(1);
         }
         fprintf(stderr, "done");
         fprintf(connected_fpout, "%d\n", 1);     // signal the client
         fflush(connected_fpout);
      }
      fprintf(stderr, "\nDone with client.\n");
      if (close(connected_fd) < 0)                // close the "connected" socket
         unix_error("close() error");
   }

   exit(0);
}//main


/*
   Below are two error handling functions.
   They provide reasonably meaningful error messages.
*/
void unix_error(char *msg)
{
   fprintf(stderr, "%s: %s (errno = %d)\n", msg, strerror(errno), errno);
   exit(1);
}

void dns_error(char *msg)
{
   fprintf(stderr, "%s: %s (h_errno = %d)\n", msg, hstrerror(h_errno), h_errno);
   exit(1);
}
