/*
    This client, along with its server, uses "message passing"
    (over a socket) to implement a simple synchronization pattern
    between two processes. The synchronization pattern is to keep
    the two processes in "lockstep" with each other. Each process
    does a sequence of "computations," where each computation can
    take a different amount of time, and the processes are
    synchronized so that they begin each successive computation
    at the exact same time.

    This client is based on the client-template.c program. The only
    part that has been changed is the section where the application
    level protocol gets implemented.
*/
#include <stdio.h>      // for fprintf() and fdopen()
#include <stdlib.h>     // for atoi() and exit()
#include <string.h>     // for memset(), bzero(), bcopy(), and strerror()
#include <errno.h>      // for errno

#include <unistd.h>     // for close()
#include <sys/socket.h> // for socket(), connect(), AF_INET, SOCK_STREAM
#include <arpa/inet.h>  // for sockaddr_in, htons()
#include <netdb.h>      // for gethostbyname(), hostent, hstrerror(), and h_errno

// prototypes for two error handling functions (defined below)
void unix_error(char *msg);
void  dns_error(char *msg);

#define ITERATION_COUNT 12

int main(int argc, char **argv)
{
   int socket_fd;                 // socket file descriptor
   struct hostent *hp;            // used to get server's IP address
   struct sockaddr_in serverAddr; // server address structure
   FILE *socket_fpout;            // socket file pointer (i.e., stream)
   FILE *socket_fpin;             // socket file pointer

   int workTime;
   int i, j;

   if (argc != 3)
   {
      fprintf(stderr, "usage: %s <host> <port>\n", argv[0]);
      exit(0);
   }

   // Create a reliable, stream socket using TCP.
   if ( (socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
      unix_error("socket() error");


   // Construct a server address structure.
   bzero(&serverAddr, sizeof(serverAddr));        // zero out the address structure
   memset(&serverAddr, 0, sizeof(serverAddr));    // zero out the address structure
   serverAddr.sin_family = AF_INET;               // Internet address family
   if ( (hp = gethostbyname( argv[1] )) == NULL ) // fill in server's IP address
      dns_error("gethostbyname() error");
   bcopy(hp->h_addr, (struct sockaddr*)&serverAddr.sin_addr, hp->h_length);
   serverAddr.sin_port = htons( atoi(argv[2]) );  // fill in the port number


   // Establish a connection with the server.
   if (connect(socket_fd, (struct sockaddr*)&serverAddr, sizeof(serverAddr)) < 0)
      unix_error("connect() error");


   // Convert the socket file descriptor to file pointers
   // (i.e., streams) and use C Standard Library I/O functions.
   if ( (socket_fpout = fdopen(socket_fd, "w")) == NULL )
      unix_error("fdopen() error" );
   if ( (socket_fpin = fdopen(socket_fd, "r")) == NULL )
      unix_error("fdopen() error" );

   srand(11);

   for (j = 0; j < ITERATION_COUNT; j++)
   {
      fprintf(socket_fpout, "%d\n", 1);      // signal the server
      fflush(socket_fpout);
      fscanf(socket_fpin, "%d", &workTime);  // wait for the server
      workTime = 1 + rand()%9;               // 1 to 9 seconds of work time
      fprintf(stderr, "\nClient: Starting a %d second calculation", workTime);
      for (i = 0; i < workTime; i++)
      {
         fprintf(stderr, ".");
         sleep(1);
      }
      fprintf(stderr, "done");
   }
   workTime = -1;                            // signal the server to stop
   fprintf(socket_fpout, "%d\n", workTime);
   fflush(socket_fpout);
   if (close(socket_fd) < 0)                 // close the socket
      unix_error("close() error");

   exit(0);
}//main


/*
   Below are two error handling functions.
   They provide reasonably meaningful error messages.
*/
void unix_error(char *msg)
{
   fprintf(stderr, "%s: %s (errno = %d)\n", msg, strerror(errno), errno);
   exit(1);
}

void dns_error(char *msg)
{
   fprintf(stderr, "%s: %s (h_errno = %d)\n", msg, hstrerror(h_errno), h_errno);
   exit(1);
}
